#!/usr/bin/env perl
# PODNAME: debrief-42.pl
# ABSTRACT: Summarize the results of a run of forty-two.

use Modern::Perl '2011';
use autodie;

use Getopt::Euclid qw(:vars);
use Smart::Comments;

use File::Basename;
use Spreadsheet::WriteExcel;

use Bio::MUST::Core;
use aliased 'Bio::MUST::Core::Ali';


my @stats;
my %org_stats_for;

for my $infile (@ARGV_infiles) {

    ### Processing: $infile
    my $ali_42 = Ali->load($infile);

    $infile =~ s/$ARGV_in_strip//xms;
    ### Comparing to: $infile
    my ($basename) = fileparse($infile, qr{\.[^.]*}xms);
    my $ali = Ali->load($infile);

    # compute global stats
    my $seqs    =    $ali->count_seqs;
    my $seqs_42 = $ali_42->count_seqs;
    my $added   = $seqs_42 - $seqs;

    # compute and store per-org stats
    my %count_for;
    $count_for{ $_->full_org }++ for $ali->all_seq_ids;

    my %count_for_42;
    $count_for_42{ $_->full_org }++ for $ali_42->all_seq_ids;

    for my $org (keys %count_for_42) {
        my $n = $count_for_42{$org} - ( $count_for{$org} // 0 );
        $org_stats_for{$org}{$basename} = $n if $n;
    }

    my $mul = grep { $count_for_42{$_} > 1 } keys %count_for_42;

    # store global stats
    push @stats, [ $basename, $seqs, $added, $mul ];
}

### Writing output file: $ARGV_outfile

# setup Excel file
my $workbook = Spreadsheet::WriteExcel->new($ARGV_outfile);
my $bf_format = $workbook->add_format;
$bf_format->set_bold;

# output global sheet
my $gbl_sheet = $workbook->add_worksheet('global');
my $line = 0;
$gbl_sheet->write_row($line++, 0, [ qw(ali seqs added mul) ], $bf_format);
$gbl_sheet->write_row($line++, 0, $_) for @stats;

# output per-org sheet
my $org_sheet = $workbook->add_worksheet('per-org');
$line = 0;
$org_sheet->write_row($line++, 0, [ qw(org ali added) ], $bf_format);
for my $org (sort keys %org_stats_for) {
    for my $basename (sort keys %{ $org_stats_for{$org} }) {
        my @fields = ( $org, $basename, $org_stats_for{$org}{$basename} );
        $org_sheet->write_row($line++, 0, \@fields);
    }
}

__END__

=pod

=head1 NAME

debrief-42.pl - Summarize the results of a run of forty-two.

=head1 VERSION

version 0.161640

=head1 USAGE

    debrief-42.pl <infiles> --in=<str> --out=<file> [optional arguments]

=head1 REQUIRED ARGUMENTS

=over

=item <infiles>

Path to input ALI files [repeatable argument].

debrief-42 should not be called in a shell loop. Use shell jokers instead:

    debrief-42.pl --main-out=42-stats.xls --in=-42 test/*-42.ali

=for Euclid: infiles.type: readable
    repeatable

=item --in[-strip]=<str>

Substring to strip from 42-built infile filenames to derive pre-42 filenames.

=for Euclid: str.type: string

=item --out[file]=<file>

Path to output .xls file summarizing global and per-org statistics.

=for Euclid: file.type: writable

=back

=head1 OPTIONAL ARGUMENTS

=over

=item --version

=item --usage

=item --help

=item --man

Print the usual program information

=back

=head1 AUTHOR

Denis BAURAIN <denis.baurain@ulg.ac.be>

=head1 COPYRIGHT AND LICENSE

This software is copyright (c) 2013 by University of Liege / Unit of Eukaryotic Phylogenomics / Denis BAURAIN.

This is free software; you can redistribute it and/or modify it under
the same terms as the Perl 5 programming language system itself.

=cut
