#!/usr/bin/env perl
# PODNAME: ompa-pa.pl
# ABSTRACT: Extract seqs from BLAST/HMMER interactively or in batch mode

use Modern::Perl '2011';

use Getopt::Euclid qw(:vars);
use IO::Prompt::Simple;
use Smart::Comments;

use Bio::MUST::Core;
use aliased 'Bio::MUST::Core::Ali::Stash';
use aliased 'Bio::MUST::Tools::OmpaPa::Blast';
use aliased 'Bio::MUST::Tools::OmpaPa::Hmmer';

# TODO: re-implement BLAST parsing and nr extraction ; done?

# setup OmpaPa sub-class based on report type
my $class = $ARGV_report_type eq 'blastxml' ? Blast : Hmmer;

# build args hash for Bio::MUST::Tools::OmpaPa::XXX constructor
my %args;
$args{database} = $ARGV_database;
$args{extract_seqs} = 1 if $ARGV_extract_seqs;

# process infiles
for my $infile (@ARGV_infiles) {
    say "### Processing <$infile>...";

    # read and parse report
    my $oum = $class->new( file => $infile, %args );
    say '[' . $oum->count_hits . ' hits processed] ';

    # BATCH MODE
    if ($ARGV_restore_bounds) {
        $oum->restore_bounds;
        say q{Here's the selection based on restored bounds...};
        $oum->list_selection;
        say '[' . $oum->selection . ' hits selected] ';
    }

    # INTERACTIVE MODE
    else {

        # repeat until user's confirmation of selected bounds
        my $ans;
        do {
            $oum->select_bounds;
            say q{Here's your current selection...};
            $oum->list_selection;
            say '[' . $oum->selection . ' hits selected] ';
            $ans = prompt 'Are you satistfied with your selection?', 'Y';
        } until uc($ans) eq 'Y';
    }

    # write accession file corresponding to current selection
    $oum->save_selection;
}

__END__

=pod

=head1 NAME

ompa-pa.pl - Extract seqs from BLAST/HMMER interactively or in batch mode

=head1 VERSION

version 0.161640

=head1 USAGE

    ompa-pa.pl <infiles> --database=<file> [optional arguments]

=head1 REQUIRED ARGUMENTS

=over

=item <infiles>

Path to input BLAST/HMMER report files [repeatable argument].

=for Euclid: infiles.type: readable
    repeatable

=item --database=<file>

Path to the sequence database used to generate the reports. For efficiency,
this argument must always be the basename of a BLAST database, even when the
reports where obtained using C<hmmsearch> on a FASTA file.

To build such a database, use one of the following commands:

    $ makeblastdb -in database.fasta -out database -dbtype prot -parse_seqids
    $ makeblastdb -in database.fasta -out database -dbtype nucl -parse_seqids

=for Euclid: file.type: string

=back

=head1 OPTIONAL ARGUMENTS

=over

=item --report-type=<str>

Type of the reports used as infiles [default: blastxml]. Currently, the
following types are available:

    - blastxml (XML BLAST reports generated with -outfmt 5)
    - hmmertbl (tabular HMMER reports generated with -tblout)

=for Euclid: str.type:       /blastxml|hmmertbl/
    str.type.error: <str> must be one of blastxml or hmmertbl (not str)
    str.default:    'blastxml'

=item --extract-seqs

Extract-seq switch [default: no]. When specified, selected sequences are
stored into a FASTA file using the same basename as other output files.

=item --restore-bounds

Batch-mode switch [default: no]. When specified, bounding-boxes are restored
from the corresponding bounds files.

=item --version

=item --usage

=item --help

=item --man

Print the usual program information

=back

=head1 AUTHOR

Denis BAURAIN <denis.baurain@ulg.ac.be>

=head1 COPYRIGHT AND LICENSE

This software is copyright (c) 2013 by University of Liege / Unit of Eukaryotic Phylogenomics / Denis BAURAIN.

This is free software; you can redistribute it and/or modify it under
the same terms as the Perl 5 programming language system itself.

=cut
