package Bio::MUST::Tools::FortyTwo::AliProcessor;
# ABSTRACT: internal class for forty-two tool
$Bio::MUST::Tools::FortyTwo::AliProcessor::VERSION = '0.161640';
use Moose;
use namespace::autoclean;

use autodie;
use feature qw(say);

use Smart::Comments -ENV;

use Carp;
use Const::Fast;
use List::AllUtils qw(uniq);
use POSIX;

use Bio::MUST::Core;
use Bio::MUST::Core::Utils qw(:filenames secure_outfile);
use aliased 'Bio::MUST::Core::Ali';
use aliased 'Bio::MUST::Core::IdList';
use aliased 'Bio::MUST::Drivers::Blast::Database::Temporary';
use aliased 'Bio::MUST::Drivers::Blast::Query';
use aliased 'Bio::MUST::Tools::SlaveAligner::Local';
use aliased 'Bio::MUST::Tools::FortyTwo::OrgProcessor';


has 'run_proc' => (
    is       => 'ro',
    isa      => 'Bio::MUST::Tools::FortyTwo::RunProcessor',
    required => 1,
);


has 'ali' => (
    is       => 'ro',
    isa      => 'Bio::MUST::Core::Ali',
    required => 1,
    coerce   => 1,
);


has 'lookup' => (
    is       => 'ro',
    isa      => 'Bio::MUST::Core::IdList',
    init_arg => undef,
    lazy     => 1,
    builder  => '_build_lookup',
);


has 'blastdb' => (
    is       => 'ro',
    isa      => 'Bio::MUST::Drivers::Blast::Database::Temporary',
    init_arg => undef,
    lazy     => 1,
    builder  => '_build_blastdb',
);


has 'integrator' => (
    is       => 'ro',
    isa      => 'Bio::MUST::Tools::SlaveAligner::Local',
    init_arg => undef,
    lazy     => 1,
    builder  => '_build_integrator',
);


has 'tax_report' => (
    is       => 'ro',
    isa      => 'Maybe[FileHandle]',
    init_arg => undef,
    writer   => '_set_tax_report',
);


has '_' . $_ . '_seqs_by_org' => (
    traits   => ['Hash'],
    is       => 'ro',
    isa      => 'HashRef[ArrayRef[Bio::MUST::Core::Seq]]',
    init_arg => undef,
    lazy     => 1,
    builder  => '_build_' . $_ . '_seqs_by_org',
    handles  => {
        'all_' . $_ . '_orgs' => 'keys',
             $_ . '_seqs_for' => 'get',
    },
) for qw(ali non);


has 'query_seqs' => (
    is       => 'ro',
    isa      => 'Bio::MUST::Drivers::Blast::Query',
    init_arg => undef,
    lazy     => 1,
    builder  => '_build_query_seqs',
);


has '_best_hits_by_ref_org' => (
    traits   => ['Hash'],
    is       => 'ro',
    isa      => 'HashRef[Bio::MUST::Core::IdList]',
    init_arg => undef,
    lazy     => 1,
    builder  => '_build_best_hits_by_ref_org',
    handles  => {
      count_ref_orgs      => 'count',
        all_ref_orgs      => 'keys',
     remove_ref_org       => 'delete',
        all_best_hits     => 'values',
            best_hits_for => 'get',
    },
);


has '_seqs_for_removal' => (
    is       => 'ro',
    isa      => 'Bio::MUST::Core::IdList',
    init_arg => undef,
    default  => sub { IdList->new() },
    handles => {
        mark_seq_for_removal => 'add_id',
             all_marked_seqs => 'all_ids',
          remove_marked_seqs => 'negative_list',
    },
);


## no critic (ProhibitUnusedPrivateSubroutines)

sub _build_lookup {
    return shift->ali->new_lookup;
}


sub _build_blastdb {
    return Temporary->new( seqs => shift->ali );
}


sub _build_integrator {
    return Local->new( ali => shift->ali );
}


sub _build_ali_seqs_by_org {
    my $self = shift;
    return $self->_split_seqs_by_org($self->ali);
}


sub _build_non_seqs_by_org {
    my $self = shift;

    my $nonfile = change_suffix($self->ali->filename, '.non');
    unless (-e $nonfile) {
        carp "Cannot open $nonfile; returning empty set of seqs!";
        return {};
    }

    return $self->_split_seqs_by_org( Ali->load($nonfile) );
}


const my $DEF_ORG => ':default';

sub _split_seqs_by_org {
    my $self = shift;
    my $seqs = shift;               # actually an Ali object...

    # split Seqs into one or more ArrayRefs keyed by full_org
    my %seqs_for;
    for my $seq ($seqs->all_seqs) {
        my $genus = $seq->genus     // q{};
        my $org   = $seq->full_org  // $DEF_ORG;
        push @{ $seqs_for{$org} }, $seq;
    }

    return \%seqs_for;
}


sub get_longest_query_seq {
    my $self = shift;

    my $max_seq;
    my $max_len = 0;

    # loop through potential queries and select longest seq

    ORG:
    for my $org ($self->all_ali_orgs) {

        SEQ:
        for my $seq ( @{ $self->ali_seqs_for($org) } ) {

            # filter out doubtful seqs from each query_org
            next SEQ if $seq->is_doubtful;

            # find longest seq using a manual yet fast approach
            my $len = $seq->nomiss_seq_len;
            if ($len > $max_len) {
                $max_seq = $seq;
                $max_len = $len;
            }
        }
    }

    return $max_seq;
}


sub _build_query_seqs {
    my $self = shift;

    #### [ALI] Collecting queries...

    my $run_proc = $self->run_proc;

    my @seqs    # filter out doubtful seqs from each query_org
        = grep { not $_->is_doubtful }
           map { @{ $self->ali_seqs_for($_) // [] } }
        $run_proc->all_query_orgs
    ;

    unless (@seqs) {
        carp 'Warning: no seq for any query_org; using longest seq instead!';
        @seqs = ( $self->get_longest_query_seq );
    }

    return Query->new( seqs => \@seqs );
}


sub _build_best_hits_by_ref_org {
    my $self = shift;

    #### [ALI] Identifying best hits for queries in reference orgs...

    my $run_proc = $self->run_proc;
    my $args = $run_proc->blast_args_for('references') // {};
    $args->{-outfmt} = 6;
    $args->{-max_target_seqs} = 10;

    my %avg_score_for;
    my %best_hits_for;

    ORG:
    for my $ref_org ($run_proc->all_ref_orgs) {

        my $query_seqs = $self->query_seqs;
        my $blastdb = $run_proc->ref_blastdb_for($ref_org);
        my $parser = $query_seqs->blastp($blastdb, $args);
        #### [ALI] BLASTP: $ref_org . q{ } . $parser->filename

        # collect best hit(s) for each query against ref proteome
        my $curr_query = q{};
        my $seen_qry_n = 0;
        my $seen_hit_n = 0;
        my $best_score = 0;
        my $cmul_score = 0;

        my %hits;
        while (my $hsp = $parser->next_hit) {

            # fetch query and bitscore for current hit
            my $query = $hsp->query_id;
            my $score = $hsp->bit_score;

            # reset score if next query
            if ($query ne $curr_query) {
                $seen_qry_n++;
                $curr_query = $query;
                $best_score = 0;
            }

            # include hit if within score tolerance
            # Note: for consistency with coverage_mul the new lower score
            # becomes the new 'best_score' (increasing tolerance thus)
            if ($score >= $best_score * $run_proc->bitscore_mul) {
                my $hit = $hsp->hit_id;
                ######## [DEBUG] query: $query_seqs->long_id_for($query)
                ######## [DEBUG] hit: $hit
                ######## [DEBUG] score: $score
                $hits{$hit}++;
                $seen_hit_n++;
                $best_score  = $score;
                $cmul_score += $score;
            }
        }
        $parser->remove unless $run_proc->debug_mode;

        # skip ref_orgs without best hits
        unless ($cmul_score) {
            carp "No best hit for: $ref_org; removing it from ref_orgs!";
            next ORG;
        }

        # store average score for best hits in current ref proteome
        # Note: divisor is set to number of best hits (plus unseen queries)
        my $divisor = $seen_hit_n + $query_seqs->count_seqs - $seen_qry_n;
        my $avg_score = $cmul_score / $divisor;
        $avg_score_for{$ref_org} = $avg_score;

        # store best hit list for current ref proteome
        my $best_hits = IdList->new( ids => [ keys %hits ] );
        $best_hits_for{$ref_org} = $best_hits;
    }
    #### [ALI] avg scores: %avg_score_for

    my $ref_org_n = ceil( $run_proc->ref_org_mul * $run_proc->all_ref_orgs );
    my @ref_orgs = keys %avg_score_for;
    if (@ref_orgs > $ref_org_n) {

        #### [ALI] Discarding non-best ref_orgs and keeping only: $ref_org_n
        @ref_orgs = sort {
            $avg_score_for{$b} <=> $avg_score_for{$a}
        } @ref_orgs;
        @ref_orgs = @ref_orgs[ 0 .. $ref_org_n - 1 ];

        # TODO: use a Tie::IxHash here to have ordered ref_orgs?
        %best_hits_for = map { $_ => $best_hits_for{$_} } @ref_orgs;
    }

    return \%best_hits_for;
}

## use critic


sub _check_brh_among_best_hits {
    my $self = shift;

    #### [ALI] Checking BRH among best hits in reference orgs...
    if ($self->count_ref_orgs < 2) {
        carp 'Warning: not enough ref_orgs for checking BRH among best hits';
        return;
    }

    my $run_proc = $self->run_proc;
    my $args = $run_proc->blast_args_for('references') // {};
    $args->{-outfmt} = 6;
    $args->{-max_target_seqs} = 1;

    for my $ref_org1 ($self->all_ref_orgs) {

        # build query_seqs from best hits for current ref_org
        my $query_seqs = Query->new(
            seqs => $run_proc->ref_blastdb_for($ref_org1)->blastdbcmd(
                [ $self->best_hits_for($ref_org1)->all_ids ]
            )
        );

        my %count_for;

        ORG2:
        for my $ref_org2 ($self->all_ref_orgs) {
            next ORG2 if $ref_org2 eq $ref_org1;

            my $blastdb = $run_proc->ref_blastdb_for($ref_org2);
            my $parser = $query_seqs->blastp($blastdb, $args);
            #### [ALI] BLASTP: "$ref_org1 vs $ref_org2 " . $parser->filename

            my $best_hits = $self->best_hits_for($ref_org2);
            while (my $hsp = $parser->next_query) {
                my $candidate = $query_seqs->long_id_for( $hsp->query_id );
                my $in_best_hits = $best_hits->is_listed( $hsp->hit_id   );
                ######## [DEBUG]: $candidate . ( $in_best_hits && ' [=BRH=]' )
                $count_for{$candidate}++ if $in_best_hits;
            }
            $parser->remove unless $run_proc->debug_mode;
        }
        #### [ALI] BRH counts: %count_for

        # remove ref_orgs that completely failed BRH
        # Note: this should not happen too often because of ref_org_mul
        unless (keys %count_for) {
            carp "Failed BRH for: $ref_org1; removing it from ref_orgs!";
            $self->remove_ref_org($ref_org1);
        }

        # check BRH with all *other* ref_orgs (hence - 1)
        my $other_n = $self->all_ref_orgs - 1;
        carp 'Warning: best hits not in BRH across all ref_orgs!'
            unless List::AllUtils::all {
                $count_for{$_} == $other_n
            } keys %count_for
        ;
    }

    return;
}


sub display {                               ## no critic (RequireArgUnpacking)
    return join "\n=== ", q{}, @_
}


sub BUILD {
    my $self = shift;

    my $run_proc = $self->run_proc;

    my $ali = $self->ali;
    #### [ALI] #seqs: $ali->count_seqs
    unless ($ali->count_seqs) {
        #### [ALI] empty file; skipping!
        return;
    }

    #### [ALI] queries: display( $self->query_seqs->all_long_ids )

    unless ($run_proc->brh_mode eq 'disabled') {
        #### [ALI] reference orgs: display( $self->all_ref_orgs )

        #### [ALI] best hits: display( map { $_->all_ids } $self->all_best_hits )

        $self->_check_brh_among_best_hits;
    }

    my $filename = $ali->filename;
    my $suffix = $run_proc->out_suffix;

    if ($run_proc->tax_reports) {
        my $tax_report = secure_outfile(
            change_suffix($filename, '.tax-report'), $suffix
        );
        open my $fh, '>', $tax_report;
        $self->_set_tax_report($fh);
        say {$fh} '# ' . join "\t", qw(
            seq_id contam_org
            rel_n mean_len mean_ident lca lineage
            acc start end strand
            seq
        );
    }

    for my $org ($run_proc->all_orgs) {
        #### [ALI] Processing ORG: $org->{org}
        OrgProcessor->new( ali_proc => $self, %{$org} );
    }

    #### [ALI] Making delayed indels...
    $self->integrator->make_indels;

    #### [ALI] Removing lengthened seqs...

    #### [ALI] lengthened seqs: display( $self->all_marked_seqs )
    $ali->apply_list( $self->remove_marked_seqs($ali) );

    #### [ALI] Writing updated file...
    my $outfile = secure_outfile($filename, $suffix);
    $ali->store($outfile);

    return;
}


__PACKAGE__->meta->make_immutable;
1;

__END__

=pod

=head1 NAME

Bio::MUST::Tools::FortyTwo::AliProcessor - internal class for forty-two tool

=head1 VERSION

version 0.161640

=head1 AUTHOR

Denis BAURAIN <denis.baurain@ulg.ac.be>

=head1 COPYRIGHT AND LICENSE

This software is copyright (c) 2013 by University of Liege / Unit of Eukaryotic Phylogenomics / Denis BAURAIN.

This is free software; you can redistribute it and/or modify it under
the same terms as the Perl 5 programming language system itself.

=cut
