package Bio::MUST::Tools::FortyTwo::OrgProcessor;
# ABSTRACT: internal class for forty-two tool
$Bio::MUST::Tools::FortyTwo::OrgProcessor::VERSION = '0.161640';
use Moose;
use namespace::autoclean;

use autodie;
use feature qw(say);

use Smart::Comments -ENV;

use Carp;
use Const::Fast;
use List::AllUtils qw(each_array);
use Path::Class qw(file);
use Tie::IxHash;

use Bio::MUST::Core;
use Bio::MUST::Core::Constants qw(:gaps);
use Bio::MUST::Core::Utils qw(:filenames);
use aliased 'Bio::MUST::Core::Ali';
use aliased 'Bio::MUST::Core::IdList';
use aliased 'Bio::MUST::Core::Seq';
use aliased 'Bio::MUST::Core::SeqId';
use aliased 'Bio::MUST::Core::SeqMask';
use aliased 'Bio::MUST::Core::Taxonomy::Filter';
use aliased 'Bio::MUST::Drivers::Blast::Database';
use aliased 'Bio::MUST::Drivers::Blast::Query';
use aliased 'Bio::MUST::Drivers::Exonerate::Aligned';


has 'ali_proc' => (
    is       => 'ro',
    isa      => 'Bio::MUST::Tools::FortyTwo::AliProcessor',
    required => 1,
);


has 'org' => (
    is       => 'ro',
    isa      => 'Str',
    required => 1,
);


has 'banks' => (
    traits   => ['Array'],
    is       => 'ro',
    isa      => 'ArrayRef[Str]',
    required => 1,
    handles  => {
        all_banks => 'elements',
    },
);


# TOOD: remove this and use auto-detection?
has 'bank_type' => (
    is       => 'ro',
    isa      => 'Str',
    default  => 'nucl',
);


has 'code' => (
    is       => 'ro',
    isa      => 'Str',
    default  => '1',
);


has 'tax_filter' => (
    is       => 'ro',
    isa      => 'Maybe[ArrayRef[Str]]',
);


has '_tax_filter' => (
    is       => 'ro',
    isa      => 'Maybe[Bio::MUST::Core::Taxonomy::Filter]',
    init_arg => undef,
    writer   => '_set_tax_filter',
    handles  => [ qw(is_allowed) ],
);


has $_ . '_seqs' => (
    is       => 'ro',
    isa      => 'Bio::MUST::Drivers::Blast::Query',
    init_arg => undef,
    lazy     => 1,
    builder  => '_build_' . $_ . '_seqs',
) for qw(homologous orthologous);


has 'aligned_seqs' => (
    is       => 'ro',
    isa      => 'Bio::MUST::Core::Ali',
    init_arg => undef,
    lazy     => 1,
    builder  => '_build_aligned_seqs',
);


## no critic (ProhibitUnusedPrivateSubroutines)

sub _build_homologous_seqs {
    my $self = shift;

    ##### [ORG] Searching for homologues...
    ##### [ORG] bank_type: $self->bank_type

    my $ali_proc = $self->ali_proc;
    my $run_proc = $ali_proc->run_proc;

    my $args = $run_proc->blast_args_for('homologues') // {};
    $args->{-outfmt} = 6;
    $args->{-max_target_seqs} = 10000       # TODO: generalize this idiom?
        unless defined $args->{-max_target_seqs};
    $args->{-db_gencode} = $self->code
        if $ali_proc->query_seqs->type eq 'prot' && $self->bank_type eq 'nucl';

    my @seqs;
    for my $bank ($self->all_banks) {

        ###### [ORG] Processing BANK: $bank

        my $file = file( $run_proc->bank_dir, $bank );
        my $blastdb = Database->new( file => $file );
        my $parser = $ali_proc->query_seqs->blast($blastdb, $args);
        ###### [ORG] TBLASTN (or BLASTP/N): $bank . q{ } . $parser->filename

        # collect hit ids, ranges and strands covered by queries
        my %mask_for;
        my %strand_for;
        while (my $hsp = $parser->next_hit) {
            my $mask = ( $mask_for{ $hsp->hit_id } //= SeqMask->empty_mask );
               $mask->mark_block($hsp->hit_start, $hsp->hit_end);
            $strand_for{ $hsp->hit_id } += $hsp->hit_strand;
        }
        $parser->remove unless $run_proc->debug_mode;
        # TODO: uniformize BLAST parser method names

        # TODO: check for need to skip distant HSPs as in 1331?
        # TODO: add small margin to ranges for better start and end?

        # fetch hit seqs
        my @hit_ids = keys %mask_for;
        my @hits = $blastdb->blastdbcmd( \@hit_ids )->all_seqs;

        # optionally trim hit seqs to their max range
        if ( $run_proc->trim_at_step('homologues') ) {
            my @ranges = map {
                [ $_->first_site, $_->last_site ]
            } @mask_for{ @hit_ids };
            ######## [DEBUG] ranges: display( map { join '-', @$_ } @ranges )

            my @strands = map {
                $_ >= 0 ? 1 : -1        # canonize strand using majority-rule
            } @strand_for{ @hit_ids };
            ######## [DEBUG] strands: display( @strands )

            for my $hit (@hits) {
                my ($start, $end) = @{ shift @ranges };
                my $strand = shift @strands;
                my $new_id = join ':::', $hit->full_id, $start, $end, $strand;
                my $new_seq = $hit->edit_seq($start, $end - $start + 1);
                $hit = Seq->new(
                    seq_id => $new_id,
                    seq    => $new_seq,
                );
            }
        }
        ######## [DEBUG] (trimmed) hits: display( map { $_->seq } @hits )
        # TODO: output a FASTA-formatted query file in DEBUG mode

        # add seqs from current bank to homologous seqs
        push @seqs, @hits;
    }

    # build BLAST query file from homologous seqs
    return Query->new( seqs => \@seqs );
}


sub _build_orthologous_seqs {
    my $self = shift;

    my $ali_proc = $self->ali_proc;
    my $run_proc = $ali_proc->run_proc;

    if ($run_proc->brh_mode eq 'disabled') {
        ##### [ORG] Skipping orthology assessment!
        return $self->homologous_seqs;
    }

    ##### [ORG] Identifying orthologues among homologues...

    my $args = $run_proc->blast_args_for('orthologues') // {};
    $args->{-outfmt} = 6;
    $args->{-max_target_seqs} = 1;
    $args->{-query_gencode} = $self->code
        if $ali_proc->query_seqs->type eq 'prot' && $self->bank_type eq 'nucl';

    my %count_for;
    for my $ref_org ($ali_proc->all_ref_orgs) {

        my $blastdb = $run_proc->ref_blastdb_for($ref_org);
        my $parser = $self->homologous_seqs->blast($blastdb, $args);
        ##### [ORG] BLASTX (or BLASTP/N): $ref_org . q{ } . $parser->filename

        my $homologous_seqs = $self->homologous_seqs;
        my $best_hits = $ali_proc->best_hits_for($ref_org);
        while (my $hsp = $parser->next_query) {
            my $candidate = $homologous_seqs->long_id_for( $hsp->query_id );
            my $in_best_hits = $best_hits->is_listed(      $hsp->hit_id   );
            ######## [DEBUG]: $candidate . ( $in_best_hits && ' [=BRH=]' )
            $count_for{$candidate}++ if $in_best_hits;
        }
        $parser->remove unless $run_proc->debug_mode;
    }
    ##### [ORG] BRH counts: %count_for

    # keep only homologues that are orthologous for all effective ref_orgs
    my $ref_org_n = $ali_proc->count_ref_orgs;
    my $orthologues = IdList->new(
        ids => [ grep { $count_for{$_} == $ref_org_n } keys %count_for ]
    );
    my $seqs = $orthologues->filtered_ali( $self->homologous_seqs );

    # build BLAST query file from orthologous seqs
    return Query->new( seqs => $seqs );
}


sub _build_aligned_seqs {               ## no critic (ProhibitExcessComplexity)
    my $self = shift;

    # TODO: split this into smaller subs once reliable tests are setup

    ##### [ORG] Aligning orthologues...

    my $aligned_seqs = Ali->new();

    my $ali_proc = $self->ali_proc;
    my $run_proc = $ali_proc->run_proc;

    my $args = $run_proc->blast_args_for('templates') // {};
    $args->{-outfmt} = 5;
    $args->{-max_target_seqs} = $run_proc->tf_max_hits * 2;
    $args->{-query_gencode} = $self->code
        if $ali_proc->query_seqs->type eq 'prot' && $self->bank_type eq 'nucl';

    my $blastdb = $ali_proc->blastdb;
    my $parser = $self->orthologous_seqs->blast($blastdb, $args);
    ##### [ORG] XML BLASTX (or BLASTP/N): $parser->filename

    my $bo = $parser->blast_output;
    return Ali->new() unless $bo;

    ORTHOLOGUE:
    for my $orthologue ($bo->all_iterations) {
        unless ($orthologue->count_hits) {
            ###### [ORG] skipped orthologue due to lack of significant template
            next ORTHOLOGUE;
        }

        # fetch transcript_seq (including a nearly complete full_id)
        # ... and potentially contaminating org
        my ($transcript_seq, $contam_org)
            = $self->_fetch_and_tag_transcript_seq($orthologue);
        if ($contam_org) {
            next ORTHOLOGUE if $run_proc->tf_action eq 'remove';
        }

        if ($run_proc->aligner eq 'disabled') {

            # build new_id from transcript id
            # ... and append contaminating org (if any)
            my $new_id = $transcript_seq->full_id;
               $new_id .= "...$contam_org" if $contam_org;
               $new_id .= '#NEW#';

            # use transcript seq as the new_seq
            my $new_seq = Seq->new(
                seq_id => $new_id,
                seq    => $transcript_seq->seq
            );

            ####### [ORG] Considering: $new_seq->full_id
            if ( $self->_is_redundant($new_seq, $aligned_seqs) ) {
                ####### [ORG] rejected for redundancy
                next ORTHOLOGUE;
            }

            # add new_seq as is (without alignment)
            $aligned_seqs->add_seq($new_seq);
            next ORTHOLOGUE;
        }

        # build a template_seq list as long as query coverage improves
        # exonerate will try to use the longest template for alignment
        # while BLAST will use each hit in turn (possibly as a fall-back)
        my @templates;
        my @template_seqs;
        my $best_coverage = 0;

        TEMPLATE:
        for my $template ($orthologue->all_hits) {

            # compute query coverage by template HSPs
            my $mask = SeqMask->empty_mask( $orthologue->query_len );
            $mask->mark_block($_->query_start, $_->query_end)
                for $template->all_hsps;
            my $coverage = $mask->coverage;

            last TEMPLATE if $coverage
                < $best_coverage * $run_proc->coverage_mul;
            $best_coverage = $coverage;

            # fetch template full_id
            my $template_id = SeqId->new(
                full_id => $blastdb->long_id_for($template->def)
            );

            ###### [ORG] template: $template_id->full_id
            ###### [ORG] coverage: sprintf '%.2f', $coverage

            # optionally skip template if from same org as the orthologue
            if ($run_proc->patch_mode eq 'on') {
                ######## [DEBUG] patch_mode test
                if ($template_id->full_org eq $transcript_seq->full_org) {
                    ###### [ORG] skipped same-org template due to patch_mode
                    next TEMPLATE;
                }
            }

            # fetch and cache aligned template seq from Ali
            push @templates, $template;
            push @template_seqs, $ali_proc->ali->get_seq(
                $ali_proc->lookup->index_for($template_id->full_id)
            );
        }

        # Note: as we (optionally) skip templates, we may run out of templates
        unless (@templates) {
            ###### [ORG] skipped orthologue due to lack of suitable template
            next ORTHOLOGUE;
        }

        # flag tracking exonerate failure (for exoblast mode)
        my $failure;

        if ($run_proc->aligner =~ m/exonerate|exoblast/xms) {

            # use longest template for exonerate alignment
            my $template_seq = $template_seqs[-1];

            # align transcript on template protein and get its translation
            # Note: in exonerate the translated dna_seq is called 'target_seq'
            my $exo = Aligned->new(
                dna_seq => $transcript_seq,
                pep_seq => $template_seq->clone->degap,
                code    => $self->code,
            );

            # build new_id from transcript id and exonerate model used
            # ... and append contaminating org (if any)
            my $new_id = $transcript_seq->full_id . '.E.' . $exo->model;
               $new_id .= "...$contam_org" if $contam_org;
               $new_id .= '#NEW#';

            # build new_seq from exonerate report
            my $new_seq = $exo->target_seq;
               $new_seq->set_seq_id($new_id);

            ####### [ORG] Considering: $new_seq->full_id

            unless ($new_seq->seq_len) {
                my $fate = $run_proc->aligner eq 'exoblast'
                    ? 'will retry with BLAST' : 'discarded'
                ;
                ####### [ORG] exonerate failure: $fate
                $failure = 1;
            }

            else {

                if ( $self->_is_redundant($new_seq, $aligned_seqs) ) {
                    ####### [ORG] rejected for redundancy
                    next ORTHOLOGUE;
                }

                # align new_seq on template_seq using subject_seq as a guide
                $aligned_seqs->add_seq(
                    $ali_proc->integrator->align(
                         new_seq => $new_seq,
                         subject => $exo->query_seq,
                        template => $template_seq,
                           start => $exo->query_start,
                    )
                );

            }
        }

        if ( ($run_proc->aligner eq 'blast')
          || ($run_proc->aligner eq 'exoblast' && $failure) ) {

            TEMPLATE:
            for my $template (@templates) {

                # use each template in turn for BLAST alignment
                my $template_seq = shift @template_seqs;
                last TEMPLATE unless $template_seq;

                # sort HSPs by descending start coordinate on template
                # TODO: check that this sort is indeed needed
                my @hsps = sort {
                    $b->hit_start <=> $a->hit_start
                } $template->all_hsps;

                HSP:
                for my $hsp (@hsps) {

                    # build HSP id from transcript id and template/HSP ranks
                    # ... and append contaminating org (if any)
                    my $hsp_id = $transcript_seq->full_id
                               . '.H' . $template->num . '.' . $hsp->num;
                       $hsp_id .= "...$contam_org" if $contam_org;
                       $hsp_id .= '#NEW#';

                    # build HSP seq from BLASTX report
                   (my $hsp_seq = $hsp->qseq) =~ s{\*}{$FRAMESHIFT}xmsg;
                    my $new_seq = Seq->new(
                        seq_id => $hsp_id,
                        seq    => $hsp_seq
                    );

                    # fetch aligned template seq from HSP (= subject)
                    my $subject_seq = Seq->new(
                        seq_id => $template_seq->seq_id,
                        seq    => $hsp->hseq
                    );

                    # reverse complement seqs if template on reverse in BLASTN
                    if ($ali_proc->query_seqs->type eq 'nucl'
                                && $self->bank_type eq 'nucl'
                                && $hsp->hit_strand == -1) {
                            $new_seq =     $new_seq->reverse_complemented_seq;
                        $subject_seq = $subject_seq->reverse_complemented_seq;
                    }

                    ####### [ORG] Considering: $new_seq->full_id
                    if ( $self->_is_redundant($new_seq, $aligned_seqs) ) {
                        ####### [ORG] rejected for redundancy
                        next HSP;
                    }

                    # align new_seq on template_seq using subject_seq as a guide
                    $aligned_seqs->add_seq(
                        $ali_proc->integrator->align(
                             new_seq => $new_seq,
                             subject => $subject_seq,
                            template => $template_seq,
                               start => $hsp->hit_start,
                        )
                    );
                }

            }
        }

    }
    $parser->remove unless $run_proc->debug_mode;

    ##### [ORG] Identifying lengthened seqs...

    # Note: depending on the value of the ls_action argument seqs can only be
    # removed from newly added seqs or also from preexisting seqs
    # from the user's POV it's only a matter of keeping or removing the latter
    # we cannot remove them yet because preexisting seqs are in blastdb
    my @org_seqs = sort {
        length( $a->raw_seq ) <=> length ( $b->raw_seq )
    } $aligned_seqs->all_seqs, ( $run_proc->ls_action eq 'remove' ?
        @{ $ali_proc->ali_seqs_for($self->org) // [] } : () );
    while (my $seq = shift @org_seqs) {
        $ali_proc->mark_seq_for_removal( $seq->full_id )
            if List::AllUtils::any { $seq->is_subseq_of($_) } @org_seqs;
    }

    ##### [ORG] Ordering aligned seqs...

    # return ordered new seqs...
    # -  asc by family
    # -  asc by first (non-gap) site
    # - desc by aligned seq len
    my $list = IdList->new(
        ids => [ map { $_->full_id }
        sort {
            ($a->family // q{}) cmp ($b->family // q{})
                    ||
             $a->first_site     <=>  $b->first_site
                    ||
             $b->seq_len        <=>  $a->seq_len
        }
        $aligned_seqs->all_seqs ]
    );

    # TODO: try to keep in-paralogues from the same family separated

    return $list->reordered_ali($aligned_seqs);
}

## use critic


sub _fetch_and_tag_transcript_seq {
    my $self       = shift;
    my $orthologue = shift;

    my $ali_proc = $self->ali_proc;
    my $run_proc = $ali_proc->run_proc;

    # prebuild SeqId list for all templates
    # this is needed for family affiliation (first only)
    #        ... and for assessing template taxonomy (all)
    my $blastdb = $self->ali_proc->blastdb;
    my @templates = $orthologue->all_hits;
    my @template_ids = map {
        SeqId->new( full_id => $blastdb->long_id_for( $_->def ) )
    } @templates;

    # set family to first template family (if any)
    my $family = $template_ids[0]->family // q{};
    $family .= '-' if $family;

    # set org to current (bank) org
    my $org = $self->org;
    my $contam_org = q{};

    # fetch transcript accession (= query id) from BLAST report
    my $query_def = $orthologue->query_def;
    my $transcript_acc = $self->orthologous_seqs->long_id_for($query_def);

    # extract range and strand (if any) from transcript accession
    # Note: this mostly makes sense for genomic contigs/scaffolds
    my @fields = split /:::/xms, $transcript_acc;
    my ($strip_acc, $start, $end, $strand) = @fields;

    # set acc to transcript accession
    my ($first_chunk) = $strip_acc =~ m/^(\S+)/xms;
    my @parts = split /\|/xms, $first_chunk;
    my $acc = pop @parts;
    ###### [ORG] orthologous transcript: $acc

    # setup tax_report data
    my $lca = [ 'unclassified sequences' ];
    my $rel_n;
    my $mean_len;
    my $mean_ident;

    # optionally analyse template taxonomy
    # ... to build taxonomic reports of additions
    # ... to tag potential contaminations
    my $fh = $ali_proc->tax_report;
    if ( $fh || $self->_tax_filter ) {
        my @relatives;
        my $sum_len   = 0;
        my $sum_ident = 0;

        my $tax = $run_proc->tax;
        my $ea = each_array @templates, @template_ids;

        REL:
        while ( my ($template, $template_id) = $ea->() ) {

            # accumulate at most tf_max_hits relatives for LCA inference
            last REL if @relatives >= $run_proc->tf_max_hits;

            # ensure that match to template is good enough
            my $hsp = $template->get_hsp(0);
            my $len = $hsp->align_len;
            my $ident = $hsp->identity / $len;
            next REL if $ident < $run_proc->tf_min_ident;
            next REL if $len   < $run_proc->tf_min_len;

            # try to determine taxonomy of template
            # if possible add template to list of relatives
            my @lineage = $tax->get_taxonomy_from_seq_id($template_id);
            if (@lineage) {
                $sum_len   += $len;
                $sum_ident += $ident;
                push @relatives, $template_id;
                ######## [DEBUG] hit: $template_id->full_id
                ######## [DEBUG] length: $len
                ######## [DEBUG] identity: sprintf '%.2f', $ident
                ######## [DEBUG] lineage: join '; ', @lineage
            }
        }
        ######## [DEBUG] relatives: display( map { $_->full_id } @relatives )

        unless (@relatives) {
            ###### [ORG] skipped taxonomy due to lack of suitable relatives
        }

        else {
            # optionally infer LCA from all relatives
            if ($fh) {
                $lca = $tax->get_common_taxonomy_from_seq_ids(@relatives);
                ######## [DEBUG] lca: join '; ', @{$lca}
                $rel_n = @relatives;
                $mean_len   = sprintf '%.2f', $sum_len   / $rel_n;
                $mean_ident = sprintf '%.2f', $sum_ident / $rel_n;
            }

            # optionally tag potential contamination based on first relative
            if ($self->_tax_filter) {
                my $seq_id = shift @relatives;
                ######## [DEBUG] first relative: $seq_id->full_id
                ######## [DEBUG] pass tax_filter: $self->is_allowed($seq_id)
                unless ( $self->is_allowed($seq_id) ) {
                    $org = "c#$org";
                    ($contam_org = $seq_id->org) =~ tr/ /_/;
                    ###### [ORG] contaminated by: $contam_org
                }
            }
        }
    }

    # fetch orthologous transcript and give it nearly complete full_id
    # Note: this step is currently useless when using a BLAST aligner
    my $transcript_seq = Seq->new(
        seq_id => $family . $org . '@' . $acc,
        seq    => $self->orthologous_seqs->seqs->first_seq(
                      sub { $_->full_id eq $transcript_acc }
                  )->seq
    );

    # optionally append line to taxonomic report
    if ($fh) {
        my $taxon = $lca->[-1];
        ###### [ORG] affiliated to: $taxon
        say {$fh} join "\t", map { $_ // q{} } (
            $transcript_seq->full_id,
            $contam_org,
            $rel_n, $mean_len, $mean_ident,
            $taxon, join('; ', @{$lca}),
            $acc, $start, $end, $strand,
            $transcript_seq->seq
        );
    }

    return ($transcript_seq, $contam_org);
}


sub _is_redundant {
    my $self = shift;
    my $seq  = shift;
    my $aligned_seqs = shift;           # TODO: improve this?

    my $ali_proc = $self->ali_proc;
    my $org = $self->org;

    # skip HSPs already included in the new seqs for the org
    return 1 if List::AllUtils::any {
        $seq->is_subseq_of($_)
    } $aligned_seqs->all_seqs;

    # skip HSPs already included in Ali for the org
    return 1 if List::AllUtils::any {
        $seq->is_subseq_of($_)
    } @{ $ali_proc->ali_seqs_for($org) // [] };

    # skip HSPs already included in .non file for the org
    return 1 if List::AllUtils::any {
        $seq->is_subseq_of($_)
    } @{ $ali_proc->non_seqs_for($org) // [] };

    return 0;
}


sub display {                               ## no critic (RequireArgUnpacking)
    return join "\n--- ", q{}, @_
}


sub BUILD {
    my $self = shift;

    my $ali_proc = $self->ali_proc;
    my $run_proc = $ali_proc->run_proc;

    # optionally build _tax_filter from tax_filter (specs) and tax attrs
    # Note: cannot use builder/coercion because of the need for a tax object
    if ($self->tax_filter) {
        my $tax = $run_proc->tax;
        $self->_set_tax_filter( $tax->tax_filter( $self->tax_filter ) )
            if $tax;        # warning already issued at this stage
    }

    ##### [ORG] homologues: display( $self->homologous_seqs->all_long_ids )
    return unless $self->homologous_seqs->all_long_ids;

    ##### [ORG] orthologues: display( $self->orthologous_seqs->all_long_ids )
    return unless $self->orthologous_seqs->all_long_ids;

    ##### [ORG] aligned: display( map { $_->full_id } $self->aligned_seqs->all_seq_ids )
    return unless $self->aligned_seqs->all_seq_ids;

    ##### [ORG] Adding aligned seqs to file...
    $ali_proc->ali->add_seq( $self->aligned_seqs->all_seqs );

    return;
}


__PACKAGE__->meta->make_immutable;
1;

__END__

=pod

=head1 NAME

Bio::MUST::Tools::FortyTwo::OrgProcessor - internal class for forty-two tool

=head1 VERSION

version 0.161640

=head1 AUTHOR

Denis BAURAIN <denis.baurain@ulg.ac.be>

=head1 COPYRIGHT AND LICENSE

This software is copyright (c) 2013 by University of Liege / Unit of Eukaryotic Phylogenomics / Denis BAURAIN.

This is free software; you can redistribute it and/or modify it under
the same terms as the Perl 5 programming language system itself.

=cut
