package Bio::MUST::Tools::FortyTwo::RunProcessor;
# ABSTRACT: internal class for forty-two tool
$Bio::MUST::Tools::FortyTwo::RunProcessor::VERSION = '0.161640';
use Moose;
use namespace::autoclean;

use autodie;
use feature qw(say);

use Smart::Comments;                    # logging always enabled here

use Carp;
use List::AllUtils;
use Path::Class qw(file);

use Bio::MUST::Core;
use aliased 'Bio::MUST::Core::Taxonomy';
use aliased 'Bio::MUST::Drivers::Blast::Database';
use aliased 'Bio::MUST::Tools::FortyTwo::AliProcessor';


has 'bank_dir' => (
    is       => 'ro',
    isa      => 'Str',
    required => 1,
);


has 'ref_bank_dir' => (
    is       => 'ro',
    isa      => 'Str',
);


has 'tax_dir' => (
    is       => 'ro',
    isa      => 'Maybe[Str]',
);


has 'ref_org_mapper' => (
    is       => 'ro',
    isa      => 'Bio::MUST::Core::IdMapper',
    coerce   => 1,
    handles  => {
        ref_bank_for => 'abbr_id_for',
    },
);


has 'query_orgs' => (
    traits   => ['Array'],
    is       => 'ro',
    isa      => 'ArrayRef[Str]',
    required => 1,
    handles  => {
        'count_query_orgs' => 'count',
          'all_query_orgs' => 'elements',
    },
);


has 'ref_orgs' => (
    traits   => ['Array'],
    is       => 'ro',
    isa      => 'ArrayRef[Str]',
    default  => sub { [] },
    handles  => {
        'count_ref_orgs' => 'count',
          'all_ref_orgs' => 'elements',
    },
);


has 'blast_args' => (
    traits   => ['Hash'],
    is       => 'ro',
    isa      => 'HashRef[HashRef]',
    default  => sub { {} },
    handles  => {
        blast_args_for => 'get',
    },
);


has 'seq_trimming' => (
    traits   => ['Array'],
    is       => 'ro',
    isa      => 'ArrayRef[Str]',
    default  => sub { [] },
    handles  => {
        _trim_at_step => 'first',
    },
);

# TODO: consider using a HashRef rather than an ArrayRef for seq_trimming

sub trim_at_step {
    my $self = shift;
    my $step = shift;

    return $self->_trim_at_step( sub { m/$step/xms } );
}

has 'brh_mode' => (
    is       => 'ro',
    isa      => 'Str',
    default  => 'strict',
);


has 'ref_org_mul' => (
    is       => 'ro',
    isa      => 'Num',
    default  => 1.0,
);


has 'bitscore_mul' => (
    is       => 'ro',
    isa      => 'Num',
    default  => 1.0,
);


has 'coverage_mul' => (
    is       => 'ro',
    isa      => 'Num',
    default  => 1.1,
);


has 'patch_mode' => (
    is      => 'ro',
    isa     => 'Str',
    default => 'off',
);


has 'aligner' => (
    is       => 'ro',
    isa      => 'Str',
    default  => 'blast',
);


has 'tf_mode' => (
    is       => 'ro',
    isa      => 'Str',
    default  => 'strict',
);


has 'tf_action' => (
    is       => 'ro',
    isa      => 'Str',
    default  => 'tag',
);


has 'tf_max_hits' => (
    is       => 'ro',
    isa      => 'Num',
    default  => 5,
);


has 'tf_min_ident' => (
    is       => 'ro',
    isa      => 'Num',
    default  => 0.90,
);


has 'tf_min_len' => (
    is       => 'ro',
    isa      => 'Num',
    default  => 30,
);


has 'ls_action' => (
    is       => 'ro',
    isa      => 'Str',
    default  => 'keep',
);


has 'out_suffix' => (
    is       => 'ro',
    isa      => 'Maybe[Str]',
    default  => '-42',
);


has 'orgs' => (
    traits   => ['Array'],
    is       => 'ro',
    isa      => 'ArrayRef[HashRef]',
    required => 1,
    handles  => {
        all_orgs => 'elements',
    },
);


has 'infiles' => (
    traits   => ['Array'],
    is       => 'ro',
    isa      => 'ArrayRef[Str]',
    required => 1,
    handles  => {
          all_infiles => 'elements',
    },
);


has 'tax' => (
    is       => 'ro',
    isa      => 'Maybe[Bio::MUST::Core::Taxonomy]',
    init_arg => undef,
    lazy     => 1,
    builder  => '_build_tax',
);


has '_ref_blastdb_by_org' => (
    traits   => ['Hash'],
    is       => 'ro',
    isa      => 'HashRef[Bio::MUST::Drivers::Blast::Database]',
    init_arg => undef,
    lazy     => 1,
    builder  => '_build_ref_blastdb_by_org',
    handles  => {
        ref_blastdb_for => 'get',
    },
);


has 'tax_reports' => (
    is       => 'ro',
    isa      => 'Bool',
    default  => 0,
);


has 'debug_mode' => (
    is       => 'ro',
    isa      => 'Bool',
    default  => 0,
);


## no critic (ProhibitUnusedPrivateSubroutines)

sub _build_tax {
    my $self = shift;

    my $tax_dir = $self->tax_dir;
    unless ($tax_dir) {
        carp 'Warning: no valid tax_dir specified; ignoring tax_filter!';
        return;
    }

    my $fuzzy_map = $self->tf_mode eq 'fuzzy' ? 1 : 0;

    return Taxonomy->new( tax_dir => $tax_dir, fuzzy_map => $fuzzy_map );
}


sub _build_ref_blastdb_by_org {
    my $self = shift;

    my %ref_blastdb_for;
    for my $ref_org ($self->all_ref_orgs) {
        $ref_blastdb_for{$ref_org} = Database->new(
            file => file(
                $self->ref_bank_dir,
                $self->ref_bank_for($ref_org)
            )
        );
    }

    return \%ref_blastdb_for;
}

## use critic


sub BUILD {
    my $self = shift;

    unless ($self->brh_mode eq 'disabled') {
        croak 'Error: ref_bank_dir missing from config file; aborting'
            unless $self->ref_bank_dir;

        croak 'Error: ref_org_mapper missing from config file; aborting'
            unless $self->ref_org_mapper;

        croak 'Error: BLAST databases missing from ref_org_mapper; aborting!'
            if List::AllUtils::any {
                !defined $self->ref_bank_for($_)
            } $self->all_ref_orgs
        ;
    }

    if ($self->aligner eq 'disabled') {
        croak 'Error: cannot disable aligner with nucleotide banks; aborting!'
            if List::AllUtils::any {
                $_->{bank_type} eq 'nucl'
            } $self->all_orgs
        ;
    }

    if ($self->aligner =~ m/exonerate|exoblast/xms) {
        croak 'Error: cannot use exonerate with protein banks; aborting!'
            if List::AllUtils::any {
                $_->{bank_type} eq 'prot'
            } $self->all_orgs
        ;
    }

    for my $infile ($self->all_infiles) {
        ### [RUN] Processing ALI: $infile
        AliProcessor->new(
            run_proc => $self,
            ali      => $infile,
        );
    }

    return;
}


__PACKAGE__->meta->make_immutable;
1;

__END__

=pod

=head1 NAME

Bio::MUST::Tools::FortyTwo::RunProcessor - internal class for forty-two tool

=head1 VERSION

version 0.161640

=head1 AUTHOR

Denis BAURAIN <denis.baurain@ulg.ac.be>

=head1 COPYRIGHT AND LICENSE

This software is copyright (c) 2013 by University of Liege / Unit of Eukaryotic Phylogenomics / Denis BAURAIN.

This is free software; you can redistribute it and/or modify it under
the same terms as the Perl 5 programming language system itself.

=cut
