package Bio::MUST::Tools::Leel::OrgProcessor;
# ABSTRACT: internal class for leel tool
$Bio::MUST::Tools::Leel::OrgProcessor::VERSION = '0.161640';
use Moose;
use namespace::autoclean;

use autodie;
use feature qw(say);

use Smart::Comments -ENV;

use Carp;
use Const::Fast;
use Path::Class qw(file);

use Bio::MUST::Core;
use aliased 'Bio::MUST::Core::Ali';
use aliased 'Bio::MUST::Core::IdMapper';
use aliased 'Bio::MUST::Core::SeqId';
use aliased 'Bio::MUST::Drivers::Blast::Database';
use aliased 'Bio::MUST::Drivers::Blast::Query';
use aliased 'Bio::MUST::Drivers::Exonerate::Aligned';


has 'ali_proc' => (
    is       => 'ro',
    isa      => 'Bio::MUST::Tools::Leel::AliProcessor',
    required => 1,
);


has 'org' => (
    is       => 'ro',
    isa      => 'Str',
    required => 1,
);


has 'banks' => (
    traits   => ['Array'],
    is       => 'ro',
    isa      => 'ArrayRef[Str]',
    required => 1,
    handles  => {
        all_banks => 'elements',
    },
);


has 'code' => (
    is       => 'ro',
    isa      => 'Str',
    default  => '1',
);


has 'protein_seqs' => (
    is       => 'ro',
    isa      => 'Bio::MUST::Drivers::Blast::Query',
    init_arg => undef,
    lazy     => 1,
    builder  => '_build_protein_seqs',
    handles  => [
        qw(abbr_id_for long_id_for)
    ],
);


has 'protein_seq_mapper' => (
    is       => 'ro',
    isa      => 'Bio::MUST::Core::IdMapper',
    init_arg => undef,
    lazy     => 1,
    builder  => '_build_protein_seq_mapper',
    handles  => {
        accession_for => 'long_id_for',
    },
);


has $_ . '_seqs' => (
    is       => 'ro',
    isa      => 'Bio::MUST::Core::Ali',
    init_arg => undef,
    lazy     => 1,
    builder  => '_build_' . $_ . '_seqs',
) for qw(transcript aligned);


## no critic (ProhibitUnusedPrivateSubroutines)

sub _build_protein_seqs {
    my $self = shift;

    ##### [ORG] Collecting proteins...

    my $ali_proc = $self->ali_proc;

    my $org = $self->org;
    my $ali = $ali_proc->ali;

    my @seqs = $ali->filter_seqs( sub { $_->full_org eq $org } );
    # TODO: this might made be more efficient if needed

    carp "Warning: no seq for org $org; skipping org!"
        unless @seqs;

    return Query->new( seqs => \@seqs );
}


sub _build_protein_seq_mapper {
    my $self = shift;

    my @protein_seq_ids = $self->protein_seqs->all_seq_ids;

    # abbr_ids are protein ids turned to query_ids for BLAST report (seq1 etc)
    my @abbr_ids = map {
        $self->abbr_id_for( $_->full_id )
    } @protein_seq_ids;

    # long_ids are bare accessions extracted from protein ids (Ahyp8251 etc)
    # these protein accessions will be matched against transcript accessions
    my @long_ids = map {
        $_->accession
            =~ s{ (?:\.H\d+\.\d+) | (?:\.E\.bf) | (?:\.E\.lc) \z}{}xmsgr
    } @protein_seq_ids;

    return IdMapper->new(
        long_ids => \@long_ids,
        abbr_ids => \@abbr_ids,
    );
}

const my $MAX_SHIFT => 100000;
const my $EXTRA     => 15;

sub _build_transcript_seqs {
    my $self = shift;

    ##### [ORG] Searching for transcripts...

    my $ali_proc = $self->ali_proc;
    my $run_proc = $ali_proc->run_proc;

    my $args = $run_proc->blast_args_for('transcripts') // {};
    $args->{-outfmt} = 5;
    $args->{-max_target_seqs} = 5;

    my @seqs;
    for my $bank ($self->all_banks) {

        ###### [ORG] Processing BANK: $bank

        my $file = file( $run_proc->bank_dir, $bank );
        my $blastdb = Database->new( file => $file );
        my $parser = $self->protein_seqs->tblastn($blastdb, $args);
        ###### [ORG] XML TBLASTN: $bank . q{ } . $parser->filename

        # abort if no hit
        my $bo = $parser->blast_output;
        return Ali->new() unless $bo;

        # collect transcript ids and ranges covered by proteins
        my @protein_ids;
        my @transcript_ids;
        my @ranges;

        PROTEIN:
        for my $protein ($bo->all_iterations) {

            # get protein accession
            my $protein_acc = $self->accession_for( $protein->query_def );

            TRANSCRIPT:
            for my $transcript ($protein->all_hits) {

                # get (corresponding?) transcript accession
                my @fields = split /\|/xms, $transcript->id;
                my $transcript_acc = pop @fields;

                # check that both accessions are identical
                my $rank = $transcript->num;
                unless ($protein_acc eq $transcript_acc) {
                    ####### [ORG] ids do not match: "$protein_acc vs. $transcript_acc (hit $rank)"

                    # optionally try next transcript if accessions are different
                    next TRANSCRIPT if $run_proc->id_match_mode eq 'enforce';
                }

                # mention success only in case of previous failure
                elsif ($rank > 1)  {
                    ####### [ORG] ids now do match: "$protein_acc (hit $rank)"
                }

                # store protein/transcript pair
                push @protein_ids,    $protein->query_def;
                push @transcript_ids, $transcript->id;

                # compute coverage range for transcript
                my $hsp = $transcript->next_hsp;
                my ($start, $end) = ($hsp->hit_start, $hsp->hit_end);
                while ($hsp = $transcript->next_hsp) {

                    # skip HSPs that are too distant from current range
                    # Note: this is to avoid exonerate to takes foreover
                    # in case of suboptimal hits on a single large chromosome
                    if ( abs( $hsp->hit_start - $start ) < $MAX_SHIFT ) {
                        $start = List::AllUtils::min($start, $hsp->hit_start);
                        $end   = List::AllUtils::max($end,   $hsp->hit_end  );
                    }
                }
                push @ranges, [$start -1 , $end -1];        # 0-based coords

                # no need to examine remaining transcripts
                next PROTEIN;
            }
        }
        $parser->remove unless $run_proc->debug_mode;

        # fetch transcript seqs
        my @transcripts = $blastdb->blastdbcmd( \@transcript_ids )->all_seqs;
        ######## [DEBUG] transcripts: display( map { $_->seq } @transcripts )

        ######## [DEBUG] ranges: display( map { join '-', @$_ } @ranges )

        # post-process transcript seqs:
        for my $seq (@transcripts) {

            # 1. rename transcript after (long) id of corresponding protein
            my $abbr_id = shift @protein_ids;
            $seq->set_seq_id(
                SeqId->new( full_id => $self->long_id_for($abbr_id) )
            );

            # TODO: improve this for single chromosome source
            # ... as exonerate takes years to finish in case of multiple hits
            # 2. trim transcript to its max range (with a tolerance)
            my ($start, $end) = @{ shift @ranges };
            $start = List::AllUtils::max($start - $EXTRA, 0);
            $end   = List::AllUtils::min($end   + $EXTRA, $seq->seq_len - 1);
            $seq->_set_seq( $seq->edit_seq($start, $end - $start + 1) );
        }
        ######## [DEBUG] trimmed transcripts: display( map { $_->seq } @transcripts )

        # add transcripts from current bank to seqs
        push @seqs, @transcripts
    }

    # check that all transcripts were indeed recovered
    my $missing_n = $self->protein_seqs->count_seqs - @seqs;
    carp "Warning: Cannot find $missing_n transcript(s): skipping it (them)!"
        if $missing_n;

    # TODO: reorder transcripts following proteins?
    # TODO: this would be needed in case of multiple banks per org

    return Ali->new( seqs => \@seqs );
}


sub _build_aligned_seqs {
    my $self = shift;

    ##### [ORG] Aligning transcripts...

    my $aligned_seqs = Ali->new();

    my $ali_proc = $self->ali_proc;
    my $run_proc = $ali_proc->run_proc;

    TRANSCRIPT:
    for my $transcript_seq ( $self->transcript_seqs->all_seqs ) {
        my $protein_seq = $ali_proc->protein_for( $transcript_seq->full_id );

        # align transcript on template protein and get its translation
        # Note: in exonerate the translated dna_seq is called 'target_seq'
        my $exo = Aligned->new(
            dna_seq => $transcript_seq,
            pep_seq => $protein_seq->clone->degap,
            code    => $self->code,
        );

        # build new_seq from exonerate report
        my $new_seq = $exo->target_seq;
        unless ($new_seq->seq_len) {
            ### exonerate failure: 'discarded'
            next TRANSCRIPT;
        }

        # align cds_seq on protein_seq using new_seq and subject_seq as guides
        $aligned_seqs->add_seq(
            $ali_proc->integrator->align(
                 new_seq => $new_seq,
                 subject => $exo->query_seq,
                template => $protein_seq,
                 cds_seq => $exo->spliced_seq,
                   start => $exo->query_start,
            )
        );
    }

    return $aligned_seqs;
}

## use critic


sub display {                               ## no critic (RequireArgUnpacking)
    return join "\n--- ", q{}, @_
}


sub BUILD {
    my $self = shift;

    my $ali_proc = $self->ali_proc;

    ##### [ORG] proteins: display( $self->protein_seqs->all_long_ids )
    return unless $self->protein_seqs->all_long_ids;

    ##### [ORG] transcripts: display( map { $_->full_id } $self->transcript_seqs->all_seq_ids  )
    return unless $self->transcript_seqs->all_seq_ids;

    ##### [ORG] aligned: display( map { $_->full_id } $self->aligned_seqs->all_seq_ids )
    return unless $self->aligned_seqs->all_seq_ids;

    ##### [ORG] Adding aligned seqs to file...
    $ali_proc->cds_ali->add_seq( $self->aligned_seqs->all_seqs );

    return;
}


__PACKAGE__->meta->make_immutable;
1;

__END__

=pod

=head1 NAME

Bio::MUST::Tools::Leel::OrgProcessor - internal class for leel tool

=head1 VERSION

version 0.161640

=head1 AUTHOR

Denis BAURAIN <denis.baurain@ulg.ac.be>

=head1 COPYRIGHT AND LICENSE

This software is copyright (c) 2013 by University of Liege / Unit of Eukaryotic Phylogenomics / Denis BAURAIN.

This is free software; you can redistribute it and/or modify it under
the same terms as the Perl 5 programming language system itself.

=cut
