package Bio::MUST::Tools::OmpaPa::Roles::Parsable;
# ABSTRACT: Parsable Moose role for search report objects
$Bio::MUST::Tools::OmpaPa::Roles::Parsable::VERSION = '0.161640';
use Moose::Role;

use autodie;
use feature qw(say);

use Smart::Comments;

use Carp;
use Const::Fast;
use FileHandle;
use File::Temp;
use IO::Prompt::Simple;
use List::AllUtils qw(sum);
use Path::Class qw(file);
use POSIX;
use Text::Table;
use Template;

use Bio::MUST::Core;
use Bio::MUST::Core::Utils qw(change_suffix);
use aliased 'Bio::MUST::Core::SeqId';
use aliased 'Bio::MUST::Drivers::Blast::Database';

requires 'file', 'collect_hits';


has 'database' => (
    is       => 'ro',
    isa      => 'Bio::MUST::Core::Types::File',
    required => 1,
    coerce   => 1,
);


has 'extract_seqs' => (
    is       => 'ro',
    isa      => 'Bool',
    default  => 0,
);

# TODO: switch to coercions? what about Stash?

has '_blastdb' => (
    is       => 'ro',
    isa      => 'Bio::MUST::Drivers::Blast::Database',
    init_arg => undef,
    lazy     => 1,
    builder  => '_build_blastdb',
);


has '_hits' => (
    traits   => ['Array'],
    is       => 'ro',
    isa      => 'ArrayRef[HashRef[Any]]',
    init_arg => undef,
    lazy     => 1,
    builder  => 'collect_hits',
    handles  => {
       count_hits => 'count',
         all_hits => 'elements',
      filter_hits => 'grep',
         map_hits => 'map',
    },
);


has '_avg_len' => (          # average length of top-25% hits
    is       => 'ro',
    isa      => 'Num',
    init_arg => undef,
    lazy     => 1,
    builder  => '_build_avg_len',
);


has $_ . '_file' => (
    is       => 'ro',
    isa      => 'Path::Class::File',
    init_arg => undef,
    lazy     => 1,
    builder  => '_build_' . $_ . '_file',
) for qw(bb idl fas);


has '_bound_for' => (
    traits   => ['Hash'],
    is       => 'ro',
    isa      => 'HashRef[Int]',
    init_arg => undef,
    default  => sub { {} },
    writer   => '_set_bound_for',
    handles  => {
        get_bound => 'get',
    },
);


has '_data_handle' => (
    is       => 'ro',
    isa      => 'FileHandle',
    init_arg => undef,
    lazy     => 1,
    builder  => '_build_data_handle',
    handles  => {
        _data_file => 'filename',
    },
);


has '_plot_handle' => (
    is       => 'ro',
    isa      => 'FileHandle',
    init_arg => undef,
    lazy     => 1,
    builder  => '_build_plot_handle',
);


## no critic (ProhibitUnusedPrivateSubroutines)

sub _build_blastdb {
    return Database->new( file => shift->database );
}


sub _build_bb_file {
    return file( change_suffix( shift->file, '.bb' ) );
}


sub _build_idl_file {
    return file( change_suffix( shift->file, '.idl' ) );
}


sub _build_fas_file {
    return file( change_suffix( shift->file, '.fasta' ) );
}


sub _build_avg_len {
    my $self = shift;

    # compute average length of the top-25% hits
    my $quarter = ceil( $self->count_hits / 4 );
    my @lengths = $self->map_hits( sub { $_->{'len'} } );
    my $avg_len = sum( @lengths[ 0 .. $quarter-1 ] ) / $quarter;

    return $avg_len;
}


sub _build_data_handle {
    my $self = shift;

    # open anonymous temporary file...
    # ... and ensures immediate communication with gnuplot
    my $data_handle = File::Temp->new(SUFFIX => '.dat', UNLINK => 1);

    # write pairs of -log10(evalue) / length as data points
    for my $hit ($self->all_hits) {
        say {$data_handle} join "\t", _eval2log10( $hit->{exp} ), $hit->{len};
    }

    return $data_handle;
}


sub _build_plot_handle {
    my $self = shift;

    # open a gnuplot subprocess...
    # ... and ensures that commands are immediately processed
    open my $plot_handle, '|-', 'gnuplot';
    $plot_handle->autoflush;

    return $plot_handle;
}

## use critic


sub select_bounds {
    my $self = shift;

    # configure display to allow interactive bound selection
    # TODO: handle a dumb terminal? this requires manually handling zooming
    $self->_setup_gnuplot;

    # output help message to console
    my $msg = << 'EOT';
Define the hit bounding-box using the mouse (button 3).
If needed, reset the zoom level by pressing the 'U' key in the plot window.
Press 'H' to get help for other hot keys.
When satisfied with your selection, press 'Return' in the console
EOT
    chomp $msg;
    prompt $msg;

    # recover newly defined bounds from gnuplot subprocess
    $self->_update_bounds;

    return;
}


sub restore_bounds {
    my $self = shift;

    # restore bounds from bb_file (possibly freshly created)
    $self->_set_bound_for(
        { $self->bb_file->slurp =~ m/^ (\w+) = (\-?\d+) $/xmsg }
    );

    return
}


sub selection {
    my $self = shift;

    # ensure that a (valid) bounding-box is in use
    croak 'Error: undefined hit bounding-box'
        unless defined $self->get_bound('maxlen');

    # dynamically filter hits based on current bounds
    my @selection = $self->filter_hits( sub {
            _eval2log10( $_->{'exp'} ) >= $self->get_bound('minexp')
         && _eval2log10( $_->{'exp'} ) <= $self->get_bound('maxexp')
         &&              $_->{'len'}   >= $self->get_bound('minlen')
         &&              $_->{'len'}   <= $self->get_bound('maxlen')
    } );

    return @selection;
}


sub list_selection {
    my $self = shift;

    # setup table of selected hits
    my $table = Text::Table->new( qw(accession description length evalue) );

    # fill-in table
    $table->load( map {
        [ $_->{acc}, @{$_}{ qw(dsc len exp) } ]
    } $self->selection );

    # output table
    print $table->rule('=')
        . $table->title
        . $table->rule('-')
        . $table->body
        . $table->rule('=')
    ;

    return;
}


sub save_selection {
    my $self = shift;

    # write file of accession (or GI) numbers
    my @ids = map { $_->{acc} } $self->selection;
    $self->idl_file->spew( join("\n", @ids) . "\n" );

    # optionally write file of seqs
    if ($self->extract_seqs) {
        my $seqs = $self->fetch_seqs( \@ids );

        # fix ids returned by blastdbcmd
        # TODO: consider fixing this in Bio::MUST::Drivers or in Stash
        for my $seq_id ( $seqs->all_seq_ids ) {
            ( my $full_id = $seq_id->full_id )
                =~ s/lcl\| | \s unnamed \s protein \s product//xmsg;
            $seq_id->_set_full_id($full_id);
        }

        $seqs->store_fasta( $self->fas_file )
    }

    return;
}


sub fetch_seqs {
    my $self = shift;
    my $ids  = shift;

    return $self->_blastdb->blastdbcmd($ids);
}


sub _setup_gnuplot {
    my $self = shift;

    # TODO: allow custom X11 templates
    # fill-in command template with relevant object attributes
    my $template = <<'EOT';
reset
x = "-log10(evalue)"
y = "hit length"
set term X11 title "OmpaPa: [% report %]"
set format "%.0f"
set mouse mouseformat x . ": %3.0f | " . y . ": %4.0f"
set size square
set title "average length of top-25% hits: [% avg_len %]"
set grid x y
set xlabel x
set ylabel y
plot "[% data_file %]" using 1:2 notitle with points pt 6 ps 1
EOT
    my $tt = Template->new;
    my $vars = {
        report    => $self->file,
        avg_len   => int( $self->_avg_len ),    # no fractional length
        data_file => $self->_data_file,
    };
    my $cmds;
    $tt->process(\$template, $vars, \$cmds)
        or croak $tt->error(), "\n";

    # send completed commands to gnuplot subprocess
    print { $self->_plot_handle } $cmds;

    return;
}


sub _update_bounds {
    my $self = shift;

    # delete any pre-existing bb_file
    $self->bb_file->remove
        if -e $self->bb_file;

    # ask gnuplot to export its bounds as currently defined
    my $template = << 'EOT';
set print "[% bb_file %]"
print "minexp=" . int(GPVAL_X_MIN)
print "maxexp=" . int(GPVAL_X_MAX)
print "minlen=" . int(GPVAL_Y_MIN)
print "maxlen=" . int(GPVAL_Y_MAX)
EOT
    my $tt = Template->new;
    my $vars = { bb_file => $self->bb_file };
    my $cmds;
    $tt->process(\$template, $vars, \$cmds)
        or croak $tt->error(), "\n";
    print { $self->_plot_handle } $cmds;

    # horrible hack to wait for the bb_file to be complete
    # TODO: improve robustness and portability!
    my $shell_cmd = 'wc -l ' . $self->bb_file;
    do {
        sleep(0.2)
    } until (-e $self->bb_file && qx{$shell_cmd} =~ m/^\s*4\b/xms);

    # restore bounds from freshly created bb_file
    $self->restore_bounds;

    return;
}


# private subs

{
    const my $MAXLOG10 => 308;
    const my $LOG10    => log(10);

    sub _eval2log10 {
        my $eval = shift;

        return $eval == 0 ? $MAXLOG10
             : $eval >= 1 ? 0
             :             -log($eval)/$LOG10
        ;
    }
}

no Moose::Role;
1;

__END__

=pod

=head1 NAME

Bio::MUST::Tools::OmpaPa::Roles::Parsable - Parsable Moose role for search report objects

=head1 VERSION

version 0.161640

=head1 SYNOPSIS

    # TODO

=head1 DESCRIPTION

    # TODO

=head1 AUTHOR

Denis BAURAIN <denis.baurain@ulg.ac.be>

=head1 COPYRIGHT AND LICENSE

This software is copyright (c) 2013 by University of Liege / Unit of Eukaryotic Phylogenomics / Denis BAURAIN.

This is free software; you can redistribute it and/or modify it under
the same terms as the Perl 5 programming language system itself.

=cut
